#!/usr/bin/perl


# Contemplate Reporter (reporter.cgi) version 1.0
# Copyright 1999-2003 Type A Multimedia Network (http://www.typea.net)
# This file may be used, modified, or redistributed under the terms
# of the open-source agreement contained in docs/license.txt


# use the CGI.pm module to access CGI functionality

	use CGI;
	$cgi = new CGI;



# define some useful subroutines

	sub get_file_simple {
		my $file = $_[0]; $text = "";

		$file = absolute_path($file, "filesystem");
		open (myfile, $file) or
		$text = "Error: Can't find file $file";
		unless ($text =~ /^Error: /) {
			while (<myfile>) { $text .= "$_"; }
			close (myfile);
		}
		
		# conform all line breaks to avoid cross-platform issues
		$text =~ s/\r\n/\r/g; # change DOS to Mac
		$text =~ s/\n/\r/g; # change Unix to Mac
		
		# remove extra line breaks at the end of the file so all the languages perform splits in the same way
		$text =~ s/(.*[^\r])\r*$/$1/;
		
		return $text;
	}
	
	sub get_setting {
		my $source_text = $_[0]; my $setting = $_[1]; my $allow_override = $_[2]; my $allow_partial = $_[2]; my $value = "";
		
		# allow most preferences to be overridden through the argument string
		$setting_no_spaces = $setting;
		$setting_no_spaces =~ s/[ -]/_/g;
		$settings_used .= "$setting_no_spaces,";

		$override = $cgi->param($setting_no_spaces);
		if (($allow_override)&&($override ne "")) {
			$value = $override;
		} else {
			$source_text = "\r".$source_text;  # this makes it easy to make sure settings start on a new line
			if ($allow_partial) {  # this lets us use more verbose labels in the ini file
				if ($source_text =~ /[^\w]$setting[^:\r]*: ([^\r]*)\r*/i) {
					$value = $1;
				}
			} else {  # this ensures an exact match in the pages file
				if ($source_text =~ /[^\w]$setting: ([^\r]*)\r*/i) {
					$value = $1;
				}
			}
		}
		if ($value eq "0") { $value = 0; } # this ensures that a 0 value will evaluate false in every language
		return $value;
	}
	
	sub absolute_path { # convert a reference to a file within our site to an absolute URL
		my $path = $_[0]; my $method = $_[1];
		if ($method eq "filesystem") {
		
			$script_location_copy = $script_location;
			$path_copy = $path;
			while ($path_copy =~ /\.\.\//) {
				# this avoids results like contemplate/../templates/default.html
				$script_location_copy =~ s/[^\/]+\/$//; # remove the last directory name from the script location
				$path_copy =~ s/\.\.\///; # remove the first updelim from the filename
			}
			$path = $script_location_copy.$path_copy;

		} else {
		
			$thisserver = $ENV{"HTTP_HOST"};
			if ($request =~ /\/([^=]*\/)contemplate\//i) { $thispath = $1; }
			if ($request =~ /\/([^=]*\/)assembled\//i) { $thispath = $1; }
			
			# to do: how can we read the authentication info from this request so we can pass it on?
			$login = ""; 
			
			$path =~ s/\.\.\//http:\/\/$login$thisserver\/$thispath/i;
		
		}
		return $path;
	}

	sub random_color {
		# returns a random web-safe color definition
		@colors = ("00","33","66","99","CC","FF");
		$r = @colors[rand(5)]; #rand(5) instead of rand(6) keeps them from getting too light to read
		$g = @colors[rand(5)];
		$b = @colors[rand(5)];
		$color = "\#".$r.$g.$b;
		return $color;		
	}
	
	sub month_display { # converts 2003-05 to May 2003
		my $month = $_[0];
		@month_year = split("-", $month);
		$month_display = @months[int(@month_year[1]) - 1];
		$year_display = @month_year[0];
		return "$month_display $year_display";
	}
	
	
# set up some primary variables

	# figure out the location of the contemplate directory as shown in the browser, and use that as the basis for open commands;
	# this allows us to run the script from a symlink to the real contemplate folder, and paths will be relative to the
	# apparent, rather than the actual, location of assembler.cgi
	# this will also prevent problems if Windows calculates relative paths from the server root
	$request = ($^O =~ /Win/) ? $ENV{'SCRIPT_NAME'} : ( (index($ENV{'REQUEST_URI'}, "?") != -1) ? substr($ENV{'REQUEST_URI'}, 0, index($ENV{'REQUEST_URI'}, "?")) : $ENV{'REQUEST_URI'} ) ;
	$script_location = ($^O =~ /Win/) ? $ENV{'PATH_TRANSLATED'} : $ENV{'SCRIPT_FILENAME'} ;
	$script_location =~ s/\\/\//g;
	if ($script_location =~ /(.*\/)assembled\//i) {
		$script_location = $1."contemplate/";
	} else {
		$script_location =~ s/(.*\/)[^\/]+/$1/;
	}

	$month = $cgi->param('month');
	$section = $cgi->param('section');
	$clear = $cgi->param('clear');
	$inspector_key = $cgi->param('inspector_key');
	$inspector_value = $cgi->param('inspector_value');
	$back_label = $cgi->param('back_label');
	$realtime = $cgi->param('realtime');
		
	$output = "";
	$button = "";
	$settings_used = "";
	$hits_greatest = 1;
	$blank_marker = "--";
	$graph_width = 200;
	$my_host = $ENV{'HTTP_HOST'};
	$my_domain = ($my_host =~ /www\./) ? substr($my_host, 3, length($my_host)) : $my_host ;

	$show_performance = 0;
	$performance_times = "";
	$last_performance_time = time;
	$total_performance_time = 0;

	$data_directory = "../../contemplate/reporter/data/";
	$data_file = $data_directory.$month.".txt";
	$error_file = $data_directory.$month."-errors.txt";

	@columns = ("dates", "IP_addresses", "hosts", "pages", "user_agents", "referrers");
	@columns_errors = ("dates", "IP_addresses", "hosts", "pages", "user_agents", "referrers", "message");
	
	@sections = ("dates", "IP_addresses", "hosts", "pages", "user_agents", "referrers", "countries");
	@sections_display = ("dates", "IP addresses", "hosts", "pages", "user agents", "referrers", "countries");
	for ($i=0; $i<@sections; $i++) {
		@{@sections[$i]} = ();
	}	

	@months = ("Jan", "Feb", "Mar", "Apr", "May", "Jun", "Jul", "Aug", "Sep", "Oct", "Nov", "Dec");

	%HTTP_error_codes = (
		"404", "File not found",
		"403", "Forbidden",
		"401", "Not authorized"
	);

	%country_codes = (
		"com", "US Commercial",
		"edu", "US Educational",
		"org", "US Organization",
		"net", "US Network",
		"gov", "US Government",
		"mil", "US Military",
		"ac", "Ascension Island",
		"ad", "Andorra",
		"ae", "United Arab Emirates",
		"af", "Afghanistan",
		"ag", "Antigua and Barbuda",
		"ai", "Anguilla",
		"al", "Albania",
		"am", "Armenia",
		"an", "Netherland Antilles",
		"ao", "Angola",
		"aq", "Antarctica",
		"as", "American Samoa",
		"at", "Austria",
		"au", "Australia",
		"aw", "Aruba",
		"az", "Azerbaijan",
		"ba", "Bosnia and Herzegowina",
		"bb", "Barbados",
		"bd", "Bangladesh",
		"be", "Belgium",
		"bf", "Burkina Faso",
		"bg", "Bulgaria",
		"bh", "Bahrain",
		"bi", "Burundi",
		"bj", "Benin",
		"bj", "Bermuda",
		"bo", "Bolivia",
		"br", "Brazil",
		"bs", "Bahamas",
		"bt", "Bhutan",
		"bv", "Bouvet Island",
		"bw", "Botswana",
		"by", "Belarus",
		"bz", "Belize",
		"ca", "Canada",
		"cc", "Cocos (Keeling) Islands",
		"cd", "Zaire",
		"cf", "Central African Republic",
		"cg", "Congo",
		"ch", "Switzerland",
		"ci", "Cte d'Ivoire",
		"ck", "Cook Islands",
		"cl", "Chile",
		"cm", "Cameroon",
		"cn", "China",
		"co", "Columbia",
		"cr", "Costa Rica",
		"cs", "Czechoslovakia",
		"cu", "Cuba",
		"cv", "Cape Verde",
		"cx", "Christmas Island",
		"cy", "Cyprus",
		"cz", "Czech Republic",
		"de", "Germany",
		"dj", "Djibouti",
		"dk", "Denmark",
		"dm", "Dominica",
		"do", "Dominican Republic",
		"dz", "Algeria",
		"ec", "Ecuador",
		"ee", "Estonia",
		"eg", "Egypt",
		"eh", "Western Sahara",
		"er", "Eritrea",
		"es", "Spain",
		"et", "Ethiopia",
		"fi", "Finland",
		"fj", "Fiji",
		"fk", "Falkland Islands",
		"fm", "Micronesia",
		"fr", "France",
		"fx", "France, Metropolitan",
		"ga", "Gabon",
		"gb", "Grenada",
		"ge", "Georgia",
		"gf", "French Guiana",
		"gg", "Guernsey",
		"gh", "Ghana",
		"gi", "Gibraltar",
		"gl", "Greenland",
		"gm", "Gambia",
		"gn", "Guinea",
		"gp", "Guadelope",
		"gq", "Equatorial Guinea",
		"gr", "Greece",
		"gs", "S. Georgia/S. Sandwich Is.",
		"gt", "Guatemala",
		"gu", "Guam",
		"gw", "Guinea-Bissau",
		"gy", "Guyana",
		"hk", "Hong Kong",
		"hn", "Honduras",
		"hr", "Croatia",
		"ht", "Haiti",
		"hu", "Hungary",
		"id", "Indonesia",
		"ie", "Ireland",
		"il", "Israel",
		"im", "Isle of Man",
		"in", "India",
		"io", "British Indian Ocean Territory",
		"iq", "Iraq",
		"ir", "Iran",
		"is", "Iceland",
		"it", "Italy",
		"je", "Jersey",
		"jm", "Jamaica",
		"jo", "Jordan",
		"jp", "Japan",
		"ke", "Kenya",
		"kg", "Kyrgystan",
		"kh", "Cambodia",
		"ki", "Kiribati",
		"km", "Comoros",
		"kn", "Saint Kitts and Nevis",
		"kp", "Korea, People's Republic of",
		"kr", "Korea, Republic of",
		"kw", "Kuwait",
		"ky", "Cayman Islands",
		"kz", "Kazakhstan",
		"la", "Lao Democratic Republic",
		"lb", "Lebanon",
		"li", "Liechtenstein",
		"lk", "Sri Lanka",
		"lr", "Liberia",
		"ls", "Lethoso",
		"lt", "Lithuania",
		"lu", "Luxembourg",
		"lv", "Latvia",
		"ly", "Libyan Arab Jamahiriya",
		"ma", "Morocco",
		"mc", "Monaco",
		"md", "Moldova",
		"mg", "Madagascar",
		"mh", "Marshall Islands",
		"mk", "Macedonia",
		"ml", "Mali",
		"mm", "Myanmar",
		"mn", "Mongolia",
		"mo", "Macau",
		"mp", "Northern Mariana Islands",
		"mq", "Martinique",
		"ms", "Montserrat",
		"mt", "Malta",
		"mu", "Mauritius",
		"mv", "Maldives",
		"mw", "Malawi",
		"mx", "Mexico",
		"my", "Malaysia",
		"mz", "Mozambique",
		"na", "Namibia",
		"nc", "New Caledonia",
		"ne", "Niger",
		"nf", "Norfolk Island",
		"ng", "Nigeria",
		"ni", "Nicaragua",
		"nl", "The Netherlands",
		"no", "Norway",
		"np", "Nepal",
		"nr", "Nauru",
		"nt", "Neutral Zone",
		"nu", "Niue",
		"nz", "New Zealand",
		"om", "Oman",
		"pa", "Panama",
		"pe", "Peru",
		"pf", "French Polynesia",
		"pg", "Papua New Guinea",
		"ph", "Philippines",
		"pk", "Pakistan",
		"pl", "Poland",
		"pm", "St. Pierre and Miquelon",
		"pn", "Pitcairn",
		"pr", "Puerto Rico",
		"pt", "Portugal",
		"pw", "Palau",
		"py", "Paraguay",
		"qa", "Qatar",
		"re", "Reunion",
		"ro", "Romania",
		"ru", "Russia",
		"rw", "Rwanda",
		"sa", "Saudi Arabia",
		"sb", "Solomon Islands",
		"sd", "Sudan",
		"se", "Sweden",
		"sg", "Singapore",
		"sh", "Saint Helena",
		"si", "Slovenia",
		"sj", "Svalbard/Jan Mayen Islands",
		"sk", "Slovakia",
		"sl", "Sierra Leone",
		"sm", "San Marino",
		"sn", "Senegal",
		"so", "Somalia",
		"sr", "Surinam",
		"st", "Sao Tome and Principe",
		"su", "USSR",
		"sv", "El Salvador",
		"sy", "Syrian Arab Republic",
		"sz", "Swaziland",
		"tc", "The Turks & Caicos Islands",
		"td", "Chad",
		"tf", "French Southern Territories",
		"tg", "Togo",
		"th", "Thailand",
		"tj", "Tajikistan",
		"tk", "Tokelau",
		"tm", "Turkmenistan",
		"tn", "Tunesia",
		"to", "Tonga",
		"tp", "East Timor",
		"tr", "Turkey",
		"tt", "Trinidad and Tobago",
		"tv", "Tuvalu",
		"tw", "Taiwan",
		"tz", "Tanzania",
		"ua", "Ukraine",
		"ug", "Uganda",
		"uk", "United Kingdom",
		"um", "U.S. Minor Outlying Islands",
		"us", "United States",
		"uy", "Uruguay",
		"uz", "Uzbekistan",
		"va", "Vatican City State",
		"vc", "Saint Vincent & Grenadines",
		"ve", "Venezuela",
		"vg", "British Virgin Islands",
		"vi", "U.S. Virgin Islands",
		"vn", "Vietnam",
		"vu", "Vanuatu",
		"wf", "Wallis and Futuna Islands",
		"ws", "Samoa",
		"ye", "Yemen",
		"yt", "Mayotte",
		"yu", "Yugoslavia",
		"zm", "Zambia",
		"zr", "Zaire",
		"zw", "Zimbabwe"
	);
	
	%country_codes_reversed = reverse %country_codes;
	
	
# read in a preferences (.ini) file if one exists, and show which preferences may be overridden from the URL string

	$preferences = get_file_simple("../preferences.ini");
	if ($preferences) {
		$top_n = get_setting($preferences, "top n", 1, 1);
		$graph_image = get_setting($preferences, "graph image", 1, 1);
		$graph_color = get_setting($preferences, "graph color", 1, 1);
		$update_interval = get_setting($preferences, "real-time update interval", 1, 1);
		$maximum_column_width = get_setting($preferences, "column width", 1, 1);
		$show_inspector_rollover = get_setting($preferences, "inspector rollover", 1, 1);
		$one_big_page = get_setting($preferences, "one big page", 1, 1);
	} else {
		$top_n = "";
		$graph_color = "";
		$graph_image = "";
		$update_interval = "";
		$maximum_column_width = "";
		$show_inspector_rollover = "";
		$one_big_page = "";
	}
	

# do a bit of processing on the settings, and set some defaults in case preferences.ini or individual settings are absent

	if ($top_n eq "") { $top_n = 5; }

	if ($graph_color eq "") { $graph_color = "#0000FF"; }
	$graph_color =~ s/#//; # Flash doesn't use # to denote a hex value

	if ($graph_image eq "") { $graph_image = "../graphics/logo.jpg"; }
	$graph_image = "../".$graph_image; # make the path relative to the Flash file, which is in a subdirectory of the Contemplate directory
	$graph_image =~ s/\.\.\/\.\.\//%252E%252E%252F%252E%252E%252F/; # double-encode the up delimiter to prevent Mac IE from decoding and then rewriting it
	
	if ($update_interval eq "") { $update_interval = 30; }

	if ($maximum_column_width eq "") { $maximum_column_width = 64; }

	# this is eye candy that probably slows down page rendering somewhat
	if ($show_inspector_rollover eq "") { $show_inspector_rollover = 1; }
	$show_inspector_rollover = int($show_inspector_rollover);
	
	# this shows the full data for each section on the same page as the top $top_n
	if ($one_big_page eq "") { $one_big_page = 0; }
	$one_big_page = int($one_big_page);
	

# if requested, delete the error log

	if ($clear) {
		unlink $error_file;
	}
		

# write a generic header to prepare for text output (PHP and ASP normally don't need this)

	$header = $cgi->header;
	print $header;
	

# if called from the real-time Flash viewer, provide some data for that

	if ($realtime) {
	
		# reinitialize $data_file to use the current month rather than looking in the URL string for a month
		($second, $minute, $hour, $day, $month, $year, $weekday, $yearday, $dst_flag) = localtime(time);
		$year += 1900;
		$month = sprintf("%02d", ($month + 1) % 100); 
		$data_file = "$year-$month.txt";
		$data_file_text = $data_file;
		$data_file = $data_directory.$data_file;

		# open the log file
		$text = get_file_simple($data_file);
		
		# create an array from the lines in the file
		@text = split("\r", $text);
		
		# loop through the array and count the number of lines that contain the appropriate number of columns	
		$hits_total = 0;
		for ($j=0; $j < @text; $j++) {
			@next = split("\t", @text[$j], 500); # the last argument, in part, tells the split function not to strip trailing empty elements
			if (@next == @columns) {
				$hits_total++;
			}
		}		

		# grab just the last ten lines to pass back to the Flash viewer
		if (@text > 10) {
			@tail = splice(@text, -10);
		}
		
		# add the country code to each line	
		for ($j=0; $j < @tail; $j++) {
			@next = split("\t", $tail[$j], 500);
			$host = $next[2]; # note that this is a hard-coded reference to the host that will break if we change the format of the data file; we could use the extract/clean up/reassemble routine used in the inspector_key section below to avoid that
			if ($host =~ /\.([^\.\d]+)$/) { @tail[$j] .= "\t".@country_codes{$1}; } else { @tail[$j] .= "\t".$blank_marker; }
		}
				
		# finally, print the results out for the Flash viewer to read
		$tail = join("\n", @tail);
		$tail =~ s/ /%20/g;
		$tail =~ s/\=/%3D/g;
		$tail =~ s/\&/%26/g;
		$tail =~ s/\?/%3F/g;
		$tail =~ s/\+/%2B/g;
		$tail =~ s/\(/%28/g;
		$tail =~ s/\)/%29/g;
		$tail =~ s/\;/%3B/g;
		$tail =~ s/\[/%5B/g;
		$tail =~ s/\]/%5D/g;
		print "hits=$hits_total&data_file=$data_file_text&tail=$tail";
		
		exit;
	
	}


# set up the HTML page
	
	print "
	
		<html>
	
		<head>
		
			<title>Contemplate(tm) Web Content Manager</title>
			
			<link rel=stylesheet href=../stylesheet.css type=text/css>

			<script language=JavaScript> login = 0; </script>
			<script language=JavaScript src=../login/login.js></script>		
			<script language=JavaScript> if (!login) { location = '../../index.html'; } </script>

		</head>
		
		<body bgcolor=#FFFFFF>
				
			<p class=title>Contemplate</p>
			<p class=body><a href=../index.html>Utilities menu</a> &gt; Reporter</p>
	
			<div class=body>
	
	";
	
	if ($show_performance) {
		$this_performance_time = time;
		$performance_times .= "&nbsp;&nbsp;script and page setup - ".($this_performance_time - $last_performance_time)." seconds<br>\r";
		$total_performance_time = $total_performance_time + ($this_performance_time - $last_performance_time);
		$last_performance_time = $this_performance_time;
	}


# if an inspect key is provided, focus on a cross-section of the month's report

	if ($inspector_key) {

	
	# clean up the input a bit

		for ($i=0; $i<@sections; $i++) {
			# it's nice for users to pass in the section name as the inspect key, but the section index is more useful to us
			if ($inspector_key eq @sections[$i]) { $inspector_key_index = $i; }
		}
		if ($inspector_key_index == 6) {
			# countries is really just a part of the host value
			$inspector_key_index = 2;
		}
		

	# read the data file for the current month and save the lines that match the inspect value
	
		$text = get_file_simple($data_file);
		
		$hits_total = 0;
		@text = split("\r", $text);
		for ($j=0; $j<@text; $j++) {
			@next = split("\t", @text[$j], 500);
			if (@next == @columns) { # make sure the line has all the elements we're expecting it to have
			
				# dates have to start with value, country codes have to end with value, all others have to equal value
				$match_end = ($inspector_key eq "dates") ? ".*" : "" ;
				$match_start = ($inspector_key eq "countries") ? ".*" : "" ;
				$match_value = $match_start.$inspector_value.$match_end;
				if (@next[$inspector_key_index] =~ /^$match_value$/) {
				
					# extract the values out of the log entry
					for ($i=0; $i < @columns; $i++) {
						${"this_".@columns[$i]} = (@next[$i] ne "") ? @next[$i] : $blank_marker ;
					}				
					
					# clean up some of the values
					# - convert TLD to country code
					if ($this_hosts =~ /\.([^\.\d]+)$/) { $this_countries = @country_codes{$1}; } else { $this_countries = $blank_marker; }
						
					# then add each value to the appropriate array
					@{"line_".$hits_total} = ();
					for ($i=0; $i < @sections; $i++) {
						@{"line_".$hits_total}[$i] = ${"this_".@sections[$i]};
					}
					
					$hits_total++;
					
				}
			}
		}  
			
					
	# print the matching lines in a table
	
		$output .= "<p class=heading><a name=top>Traffic for ".month_display($month)."</a></p>\r";
		
		$inspector_key_singular = substr($inspector_key, 0, -1);
		if ($inspector_key_singular eq "IP addresse") { $inspector_key_singular = substr($inspector_key_singular, 0, -1); }
		if ($inspector_key_singular eq "countrie") { $inspector_key_singular = "country"; }
		$output .= "<b>Where $inspector_key_singular contains $inspector_value</b><br>\r";
		$output .= "<b>Total hits: $hits_total</b><p>\r\r";
		
		$output .= "&lt;&ndash; <a href=JavaScript:history.back()>$back_label</a><p>\r\r";

		$output .= "<table cellpadding=2 cellspacing=0 border=0>\r\r";
		
		$output .= "<tr bgcolor=#CCCCCC>\r";
		for ($i=0; $i < @sections; $i++) {
			$output .= "<td class=subheading>".ucfirst(@sections[$i])."</td>\r";
		}
		$output .= "</tr>\r\r";
	
		for ($j=0; $j < $hits_total; $j++) {	
			$output .= "<tr valign=top>\r";
			for ($i=0; $i < @sections; $i++) {
				$output .= "<td class=body>\r";
				$output .= @{"line_".$j}[$i]."\r";
				$output .= "</td>\r";
			}
			$output .= "</tr>\r\r";
		}
		
		$output .= "</table><p>\r\r";
	
		$output .= "&lt;&ndash; <a href=JavaScript:history.back()>$back_label</a><p>\r\r";
		
	}


# if a month is specified, display that month's report

	elsif ($month) {


	# read the data file
	
		$text = get_file_simple($data_file);

		if ($show_performance) {
			$this_performance_time = time;
			$performance_times .= "&nbsp;&nbsp;load data file - ".($this_performance_time - $last_performance_time)." seconds <br>\r";
			$total_performance_time = $total_performance_time + ($this_performance_time - $last_performance_time);
			$last_performance_time = $this_performance_time;
		}


	# split the data into lines and columns and create associative arrays

		$hits_total = 0;
		@text = split("\r", $text);
		if ($show_performance) {
			$this_performance_time = time;
			$performance_times .= "&nbsp;&nbsp;split data file - ".($this_performance_time - $last_performance_time)." seconds <br>\r";
			$total_performance_time = $total_performance_time + ($this_performance_time - $last_performance_time);
			$last_performance_time = $this_performance_time;
		}
		for ($j=0; $j<@text; $j++) {
			@next = split("\t", @text[$j], 500);
			if (@next == @columns) { # make sure the line has all the elements we're expecting it to have
			
				# extract the values out of the log entry
				for ($i=0; $i < @columns; $i++) {
					${"this_".@columns[$i]} = (@next[$i] ne "") ? @next[$i] : $blank_marker ;
				}				
				
				# clean up some of the values
				# - ignore times to group hits by day
				$this_dates = substr($this_dates, 0, index($this_dates, " "));
				# - convert TLD to country code
				if ($this_hosts =~ /\.([^\.\d]+)$/) { $this_countries = @country_codes{$1}; } else { $this_countries = $blank_marker; }
				# - filter self-referrers
				if (($this_referrers =~ /$my_domain/)||(length($this_referrers) < 3)) { $this_referrers = $blank_marker; }
					
				# then add each value to the appropriate array
				for ($i=0; $i < @sections; $i++) {
					if (exists(${@sections[$i]}{${"this_".@sections[$i]}})) { # e.g. if (exists($dates["2003-05"])) {
						${@sections[$i]}{${"this_".@sections[$i]}}++;
						$this_hits = ${@sections[$i]}{${"this_".@sections[$i]}};
						if ($this_hits > $hits_greatest) { $hits_greatest = $this_hits; }
					} else {
						%{@sections[$i]} = (%{@sections[$i]}, ${"this_".@sections[$i]} => 1);
					}
				}
					
				$hits_total++;
			
			}
		}  
		
		if ($show_performance) {
			$this_performance_time = time;
			$performance_times .= "&nbsp;&nbsp;process data file - ".($this_performance_time - $last_performance_time)." seconds <br>\r";
			$total_performance_time = $total_performance_time + ($this_performance_time - $last_performance_time);
			$last_performance_time = $this_performance_time;
		}

	
	# print the reports in a table
	
		$output .= "<p class=heading><a name=top>Traffic for ".month_display($month)."</a></p>\r\r";
		
		$output .= "<p class=subheading>Total hits: $hits_total</p><br>\r\r";
		
		$output .= "<p class=body>\r";
		$output .= "<a href=reporter.cgi>Months</a> &gt; \r";
		if (($section)&&(!$one_big_page)) {
			$output .= "<a href=reporter.cgi?month=$month>Top ".$top_n."'s</a> &gt; \r";
		} else {
			$output .= "Top ".$top_n."'s &gt; \r";
		}
		$output .= "<font color=#0000FF>All</font> ";
		for ($i=0; $i < @sections; $i++) {
			$link = ($one_big_page) ? "\#@sections[$i]" : "reporter.cgi?month=$month&section=@sections[$i]" ;
			if ($section eq @sections[$i]) {
				$output .= ucfirst(@sections_display[$i]).", ";
			} else {
				$output .= "<a href=$link>".ucfirst(@sections_display[$i])."</a>, ";
			}
		}
		if ($section eq "errors_HTTP") {
			$output .= "HTTP errors, ";
		} else {
			if ($one_big_page) {
				$output .= "<a href=\#errors_HTTP>HTTP errors</a>, ";
			} else {
				$output .= "<a href=reporter.cgi?month=$month&section=errors_HTTP>HTTP errors</a>, ";
			}
		}
		if ($section eq "errors_assembler") {
			$output .= "Assembler errors<p>";
		} else {
			if ($one_big_page) {
				$output .= "<a href=\#errors_assembler>Assembler errors</a>\r";
			} else {
				$output .= "<a href=reporter.cgi?month=$month&section=errors_assembler>Assembler errors</a>\r";
			}
		}
		$output .= "</p>\r\r";
		
		$inspector_button_ID = 0;
			
		$output .= "<table cellpadding=2 cellspacing=0 border=0>\r\r";
		
		if ($show_performance) {
			$this_performance_time = time;
			$performance_times .= "&nbsp;&nbsp;set up section links - ".($this_performance_time - $last_performance_time)." seconds <br>\r";
			$total_performance_time = $total_performance_time + ($this_performance_time - $last_performance_time);
			$last_performance_time = $this_performance_time;
		}

		$passes = 2;
		for ($h=0; $h < $passes; $h++) { # print all the tables twice, the first time showing just the top $top_n ...
		
			# ... but if we're splitting the sections up, skip the first or second pass
			if (!$one_big_page) {
				if (!$section) {
					$passes = 1; # only do the first part
				} else {
					$h = 1; # only do the second part
				}
			}
			
			for ($i=0; $i < @sections; $i++) {
				
				# skip unless we're displaying this section, or no section is specified
				if ((!$section)||($section eq @sections[$i])) {
			
					if ($h == 0) {
						$label = "Top $top_n @sections_display[$i]";
						$button_link = ($one_big_page) ? "\#@sections[$i]" : "reporter.cgi?month=$month&section=@sections[$i]" ;
						$button = "<a href=$button_link>All @sections_display[$i]</a>";
						$space_after = 18;
					} else {
						$label = "<a name='@sections[$i]'>All @sections_display[$i]</a>";
						if ($one_big_page) { $button = "<a href=\#top>Top</a>"; }
						$space_after = 36;
					}
	
					$output .= "<tr><td colspan=6 bgcolor=#CCCCCC class=subheading><img src=../graphics/spacer.gif height=2 width=1><br>$label</td></tr>\r\r";
					
					# sort the all dates list by key and everything else by value
					if ((@sections[$i] =~ /dates/)&&($h == 1)) {
						@{@sections[$i]} = sort @{@sections[$i]};
						@{@sections[$i]."_keys"} = keys %{@sections[$i]};
						@{@sections[$i]."_values"} = values %{@sections[$i]};
					} else {
						@{@sections[$i]."_keys"} = sort { ((${@sections[$i]}{$b} <=> ${@sections[$i]}{$a}) || ($a cmp $b)) } keys %{@sections[$i]};
						@{@sections[$i]."_values"} = sort { int($b) <=> int($a) } values %{@sections[$i]};
					}
					
					# for the partial list, crop to the first $top_n values
					if ($h == 0) {
						splice(@{@sections[$i]."_keys"}, $top_n);
						splice(@{@sections[$i]."_values"}, $top_n);
					}
						
					if ($show_performance) {
						$this_performance_time = time;
						$performance_times .= "&nbsp;&nbsp;sort and crop @sections[$i] data - ".($this_performance_time - $last_performance_time)." seconds <br>\r";
						$total_performance_time = $total_performance_time + ($this_performance_time - $last_performance_time);
						$last_performance_time = $this_performance_time;
					}

					# okay, now we can finally print the rows
					$this_limit = (@{@sections[$i]."_keys"} < $top_n) ? @{@sections[$i]."_keys"} : $top_n ;
					$keys_count = ($h > 0) ? @{@sections[$i]."_keys"} : $this_limit ;
					for ($j=0; $j < $keys_count; $j++) {
						
						$key = @{@sections[$i]."_keys"}[$j];
						$value = @{@sections[$i]."_values"}[$j];
						
						# prepare a link to allow inspecting an item in the list
						$inspector_key = @sections[$i];
						$inspector_value = $key;
						
						# untranslate the country code for the inspect script
						if (@sections[$i] =~ /countries/) {
							$inspector_value = ".".@country_codes_reversed{$inspector_value};
						}
						
						# do some basic URL-encoding on the inspect value
						$inspector_key =~ s/ /%20/g;
						$inspector_value =~ s/ /%20/g;
						$inspector_value =~ s/\=/%3D/g;
						$inspector_value =~ s/\&/%26/g;
						$inspector_value =~ s/\?/\\%3F/g;
						$inspector_value =~ s/\+/\\%2B/g;
						$inspector_value =~ s/\(/\\%28/g;
						$inspector_value =~ s/\)/\\%29/g;
						$inspector_value =~ s/\;/\\%3B/g;
						$inspector_value =~ s/\[/\\%5B/g;
						$inspector_value =~ s/\]/\\%5D/g;
	
						# limit the item display length to a predetermined number of characters
						$key_complete = $key;
						$key = (length($key) > $maximum_column_width) ? substr($key,0,$maximum_column_width)."..." : $key ;
						
						# make the referrer items into a link
						if (@sections[$i] =~ /referrers/) {
							$key = ($key ne $blank_marker) ? "<a href=".$key_complete." target='_new'>$key</a>" : $key ;
						}
						
						# provide a link for seeing all the fields for a particular item
						$back_label = ($section) ? "All ".$section : "Top ".$top_n."%27s" ;
						$back_label =~ s/_/ /g;
						$back_label =~ s/ /%20/g;
						if ($show_inspector_rollover) {
							$inspector_link = ($key ne $blank_marker) ? "<a href='reporter.cgi?month=$month&inspector_key=$inspector_key&inspector_value=$inspector_value&back_label=$back_label'><img src=../graphics/inspector_0.gif onmouseover='document.images.inspector_$inspector_button_ID.src=\"../graphics/inspector_1.gif\"' onmouseout='document.images.inspector_$inspector_button_ID.src=\"../graphics/inspector_0.gif\"' width=12 height=12 border=0 name=inspector_$inspector_button_ID alt='Inspect'></a>" : "&nbsp;" ;
							$inspector_button_ID++;
						} else {
							$inspector_link = ($key ne $blank_marker) ? "<a href='reporter.cgi?month=$month&inspector_key=$inspector_key&inspector_value=$inspector_value&back_label=$back_label'><img src=../graphics/inspector_0.gif width=12 height=12 border=0 alt='Inspect'></a>" : "&nbsp;" ;					
						}
						
						$output .= "<tr valign=top><td class=body>\r";
						$output .= $inspector_link."&nbsp;";
						$output .= "</td>\n<td class=body>\r";
						$output .= $key;
						$output .= "</td>\n<td align=right><img src=../graphics/spacer.gif width=12 height=1></td>\n<td class=body>\r";
						$output .= $value;
						$output .= "</td>\n<td align=right><img src=../graphics/spacer.gif width=12 height=1></td>\n<td width=$graph_width>\r";
						$output .= "<font color=".random_color()."><img src=../graphics/spacer.gif height=4 width=".(($value / $hits_greatest) * $graph_width)." border=1></font>\r";
						$output .= "</td></tr>\r\r";
					
					}
					
					# more/top buttons after listings
					$output .= "<tr><td colspan=6 class=body>$button</td></tr>\r\r";
					
					# space between rows
					if (!$section) { $output .= "<tr><td colspan=6><img src=../graphics/spacer.gif height=$space_after width=1></td></tr>\r\r"; }
				
					if ($show_performance) {
						$this_performance_time = time;
						$performance_times .= "&nbsp;&nbsp;print @sections[$i] data - ".($this_performance_time - $last_performance_time)." seconds <br>\r";
						$total_performance_time = $total_performance_time + ($this_performance_time - $last_performance_time);
						$last_performance_time = $this_performance_time;
					}

				}

			}
			
			# space between sections
			if ($h == 0) {
				if ($one_big_page) { $output .= "<tr><td colspan=6><img src=../graphics/spacer.gif height=36 width=1></td></tr>\r\r"; }
			}

		}
		
			
	# look for an error file and print those, too
		
		# skip if this is the menu and we're splitting the sections up
		if (($one_big_page)||($section)) {
			
			for ($h=0; $h<2; $h++) {
			
				$error_type = ($h == 0) ? "HTTP" : "assembler" ;
			
				# skip unless we're displaying this section, or no section is specified
				if ((!$section)||($section eq "errors_".$error_type)) {
	
					$error_list = "";

					$output .= "<tr><td colspan=6 bgcolor=#CCCCCC class=subheading><img src=../graphics/spacer.gif height=2 width=1><br><a name='errors_$error_type'>All $error_type errors</a></td></tr>\r\r";
			
					$text = get_file_simple($error_file);
					
					if (index($text, "Error: ") != 0) {
						@text = split("\r", $text);
						for ($j=0; $j<@text; $j++) {
							@next = split("\t", @text[$j], 500);
							if (@next == @columns_errors) { # make sure the line has all the elements we're expecting it to have
								
								# extract the values out of the log entry
								for ($i=0; $i < @columns_errors; $i++) {
									${"this_".@columns_errors[$i]} = (@next[$i] ne "") ? @next[$i] : $blank_marker ;
								}				
					
								# clean up some of the values
								$this_page = (length($this_pages) > ($maximum_column_width + 24)) ? substr($this_pages, 0, ($maximum_column_width + 24))."..." : $this_pages ; # limit the page display length

								if (($error_type eq "assembler")&&(substr($this_message, 0, 3) !~ /\d\d\d/)&&(substr($this_message, 0, 3) !~ /Unk/)) {
									# show assembler errors first...
									$message =~ s/,/<br>/g;
									$error_list .= "<tr><td colspan=6 class=body>$this_dates<br>$this_pages<br>$this_message<hr></td></tr>\r";
								} elsif (($error_type eq "HTTP")&&((substr($this_message, 0, 3) =~ /\d\d\d/)||(substr($this_message, 0, 3) =~ /Unk/))) {
									# ...then HTTP errors
									$this_message =~ s/[^\d]//g; # strip it down to just the error code
									$this_message = (@HTTP_error_codes{$this_message}) ? "$this_message: @HTTP_error_codes{$this_message}" : "Unknown error" ;
									$error_list .= "<tr><td colspan=6 class=body>$dates<br>$pages<br>$this_message<hr></td></tr>\r";
								}
								
							}
						}
						$output .= $error_list;
					}
						
					if ($error_list) {
						# provide a button for clearing the error log
						$output .= "<tr><td colspan=6 class=body>\r";
						$output .= "<form name=clear method=post><input type=hidden name=clear value=1><input type=hidden name=month value=$month></form>\r";
						$output .= "<script language=JavaScript>\n function clearErrors() {\r";
						$output .= "if (confirm('Are you sure you want to clear the error log for this month?')) { document.clear.submit(); }\r }\r";
						$output .= "</script>\r";
						$output .= "<a href=JavaScript:clearErrors()>Clear error log</a>\r";
						$output .= "</td></tr>\r\r";
					} else {
						$output .= "<tr><td colspan=6 class=body>$blank_marker</td></tr>\r\r";
					}
					
					# more/top buttons after listings
					$output .= "<tr><td colspan=6 class=body>$button</td></tr>";
				
					# space between sections
					if ($h == 0) {
						if ($one_big_page) { $output .= "<tr><td colspan=6><img src=../graphics/spacer.gif height=36 width=1></td></tr>"; }
					}
	
				}
	
			}

		}		
		
		
	# finish off the table, and we're done
			
		$output .= "</table><br>\r\r";
		if (!$section) {
			$output .= "<p class=body>&lt;&ndash; <a href=reporter.cgi>Months</a></p>\r\r";
		} else {
			$output .= "<p class=body>&lt;&ndash; <a href=reporter.cgi?month=$month>Top ".$top_n."'s</a></p>\r\r";
		}
		

# otherwise, build a menu of all the available data files
	
	} else {
	
	
	# get a list of all the files in the data directory
	
		@data_files = ();
		
		opendir(dir, absolute_path($data_directory, "filesystem")) or
		$output .= "<b>Error:</b> Can't read the contents of the data directory $data_directory<p>";
		@data_files = grep {/^\d{4}-\d{2}\.txt$/} readdir(dir);
		$data_files_count = @data_files;
		closedir dir;
		
		@data_files = sort {$b cmp $a} (@data_files);
		
		
	# make an associative array containing the number of hits in each file

		$hits_total = 0;
		for ($i=0; $i < $data_files_count; $i++) {
			$text = get_file_simple($data_directory.@data_files[$i]);
			@text_lines = split("\r", $text);
			$hits = @text_lines;
			if ($i == 0) {
				# get the last hit out of the first file, since we've sorted the files from newest to oldest
				@last_line = split("\t", @text_lines[$hits - 1], 500);
				$last_hit = @last_line[0];
				#$last_file = @data_files[$i]; # debug
			}	

			@data_files{@data_files[$i]} = $hits;
			$hits_total = $hits_total + $hits;
			if ($hits > $hits_greatest) { $hits_greatest = $hits; }
		}


	# print the file names and number of hits in a table
	
		$month_oldest = substr(@data_files[$data_files_count - 1], 0, -4);

		$output .= "<p class=heading>Traffic since ".month_display($month_oldest)."</p>\r\r";
		
		$output .= "<p class=subheading>Last hit: $last_hit</p>\r";
		$output .= "<p class=subheading>Total hits: $hits_total</p><br>\r\r";
		
		$output .= "<table cellpadding=2 cellspacing=0 border=0>\r\r";
		
		$output .= "<tr><td colspan=5 bgcolor=#CCCCCC class=subheading><img src=../graphics/spacer.gif height=2 width=1><br>Monthly reports</td></tr>\r\r";

		for ($i=0; $i < $data_files_count; $i++) {
			$key = @data_files[$i];
			$value = @data_files{$key};
			$month = substr($key, 0, -4);
			$output .= "<tr><td class=body>\r";
			$output .= "<a href=reporter.cgi?month=$month>$month</a>\r";  # change the last $month to $key here to show the .txt extensions in the menu
			$output .= "</td>\n<td align=right><img src=../graphics/spacer.gif width=12 height=1></td>\n<td class=body>\r";
			$output .= $value;
			$output .= "</td>\n<td align=right><img src=../graphics/spacer.gif width=12 height=1></td>\n<td width=$graph_width>\r";
			$output .= "<font color=".random_color()."><img src=../graphics/spacer.gif height=4 width=".(($value / $hits_greatest) * $graph_width)." border=1></font>\r";
			$output .= "</td></tr>\r\r";
		}
		
		$output .= "<tr><td colspan=5>&nbsp;</td></tr>\r\r";

		$output .= "<tr><td colspan=5 bgcolor=#CCCCCC class=subheading><img src=../graphics/spacer.gif height=2 width=1><br>Other views</td></tr>\r\r";

		$output .= "<tr><td class=body>\r";	
		$output .= "<a href=\"JavaScript:window.open('realtime.swf?script_extension=cgi&graph_color=".$graph_color."&graph_image=".$graph_image."&update_interval=".$update_interval."','realtime','width=550,height=500,location=no,menubar=no,directories=no,toolbar=no,scrollbars=yes,resizable=yes,status=yes')\">Real-time display</a><br>\r\r";	
		$output .= "<a href=data/>Raw logs</a><br>\r";	
		$output .= "</td></tr>\r\r";	
		
		$output .= "</table><br>\r\r";
	
	}
		
			
# print everything to the browser

	print $output;
	

# if requested, show a performance report

	if ($show_performance) {
		$this_performance_time = time;
		$performance_times .= "&nbsp;&nbsp;wrap up - ".($this_performance_time - $last_performance_time)." seconds <br>\r";
		$total_performance_time = $total_performance_time + ($this_performance_time - $last_performance_time);
		$last_performance_time = $this_performance_time;
	}

	if ($show_performance) {
		print "Language: cgi<br>\r";
		print "Host: $my_host<br>\r";
		print "Total hits: $hits_total<br>\r";
		print "Performance times:<br>$performance_times\r";
		print "Total time: $total_performance_time seconds\r";
	}

# we're done!

	print "</div>\r\r";

	print "</body></html>\r\r";

	exit;
